## File: PDL::IO::SndFile.pm
## Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
## Description: PDL I/O routines for audio file formats supported by Audio::SndFile

package PDL::IO::SndFile;
use PDL;
use Audio::SndFile;
use Carp;
use Exporter;
use strict;

##======================================================================
## Globals

our $VERSION = 0.01;

our @ISA = qw(Exporter);
our @EXPORT = qw(sfinfo readsf writesf);
our @EXPORT_OK = @EXPORT;
our @EXPORT_TAGS = (
		    'all' => [@EXPORT_OK],
		   );

our %DEFAULTS =
  (
   'samplerate' => 44100, ##-- (write): default sample rate
   'type'       => 'raw', ##-- (write): default file type
   #'type'       => 'wav', ##-- (write): default file type
   'subtype'    => undef, ##-- (write): default subtype (see _pdltype_to_sfsubtype())
  );

our %known_sf_file_types =
  (
   map {($_=>undef)}
   qw(wav aiff au raw paf svx nist voc ircam w64 mat4 mat5 pvf xi htk sds avr wavex sd2 flac caf)
  );

our %known_sf_options =
  (
   map {($_=>undef)} qw(type subtype channels endianness samplerate)
  );

##======================================================================
## Constructors etc.

## $sfio = PDL::IO::SndFile->new(%args)
##  + %args:
##    sf      => $sf,       ##-- underlying Audio::SndFile object
##    options => \%options, ##-- options for Audio::SndFile::open()
##    info    => \%info,    ##-- extended info for Audio::SndFile
sub new {
  my ($that,%args) = @_;
  return bless({
		sf => undef,
		options => {},
		info => {},
		%args,
	       }, ref($that)||$that);
}

## $sfio = $sfio->clear()
##  + clears cached data and options
sub clear {
  my $sfio = shift;
  delete($sfio->{sf});
  %{$sfio->{options}} = qw();
  %{$sfio->{info}} = qw();
  return $sfio;
}

##======================================================================
## Methods: Open/close

## $sfio_or_undef = $sfio->open($mode,$file,%opts)
## $sfio_or_undef = $sfio->open("${mode}${file}",%opts)
sub open {
  my $sfio = shift;
  my $mode = shift;
  my ($file,%opts);
  if (@_ % 2 == 0) {
    ##-- called as open("${mode}${file}", %opts)
    if ($mode =~ m/^(\+?(?:\>|\>\>?))(.*)$/) {
      ($mode,$file) = ($1,$2);
    } else {
      $file = $mode;
      $mode = '<'; ##-- default mode: read only
    }
    %opts = @_;
  } else {
    ($file,%opts) = @_;
  }
  %opts = (%{$sfio->{options}}, %opts) if ($sfio->{options});
  delete(@opts{grep {!exists($known_sf_options{$_})} keys(%opts)});

  ##-- (re-)open
  $sfio->clear();
  $sfio->{sf} = Audio::SndFile->open($mode,$file,%opts)
    or confess(__PACKAGE__,"::open(): open failed for file '$file'");
  return $sfio;
}

## $sfio_or_undef = $sfio->close()
sub close {
  my $sfio = shift;
  $sfio->{sf}->close if (UNIVERSAL::can($sfio->{sf},'close'));
  $sfio->clear();
  return $sfio;
}

##======================================================================
## Methods: Info

## \%options = $sfio->options()
##  + reads options from $sfio->{sf}
sub options {
  my $sfio = shift;
  my $sf = $sfio->{sf};
  return {
	  map {$_ => ($sf ? $sf->can($_)->($sf) : undef)}
	  qw(type subtype endianness channels samplerate)
	 };
}

## \%info = $sfio->info()
##  + reads extended info from $sfio->{sf}
sub info {
  my $sfio = shift;
  my $sf = $sfio->{sf};
  my $i = {
	   %{$sfio->options},
	   map {$_ => ($sf ? $sf->can($_)->($sf) : undef)}
	   (
	    qw(sections seekable frames),
	    qw(title copyright software artist comment date)
	   )
	  };
  delete(@$i{grep {!defined($i->{$_})} keys(%$i)});
  return $i;
}

## \%info = PACKAGE::sfinfo($filename,%options)
sub sfinfo {
  my $filename = shift;
  my $sfio = __PACKAGE__->new();
  $sfio->open('<',$filename,@_)
    or confess(__PACKAGE__."::sfinfo(): open failed for file '$filename': $!");
  return $sfio->info();
}

##======================================================================
## Utilities: dispatch to underlying Audio::SndFile object

## \&wrapped = CLASS::sfwrapper($func)
sub sfwrapper {
  my $func = shift;
  return sub {
    my $sfio = shift;
    confess(ref($sfio)."::sfwrapper($func): no underlying Audio::SndFile object present!")
      if (!$sfio->{sf});
    return $sfio->{sf}->can($func)->($sfio->{sf},@_);
  };
}

## ? = $sfio->sfcall($func,@args);
sub sfcall {
  return sfwrapper($_[1])->($_[0],@_[2..$#_]);
}

##======================================================================
## Utilities: PDL <-> SndFile

## %s_nbits_to_ptype : signed integer types: ($nbits=>$pdltype)
## %u_nbits_to_ptype : unsigned integer types: ($nbits=>$pdltype)
our %s_nbits_to_ptype = (map { 8*PDL::howbig($_)=>$_ } (PDL::byte,PDL::short,PDL::long,PDL::longlong));
our %u_nbits_to_ptype = (map { 8*PDL::howbig($_)=>$_ } (PDL::byte,PDL::ushort,PDL::long,PDL::longlong));

## %s_ptype_to_nbits : signed integer types: ($pdltype=>$nbits)
## %u_ptype_to_nbits : unsigned integer types: ($pdltype=>$nbits)
our %s_ptype_to_nbits = (map { $_=>8*PDL::howbig($_) } (PDL::byte,PDL::short,PDL::long,PDL::longlong));
our %u_ptype_to_nbits = (map { $_=>8*PDL::howbig($_) } (PDL::byte,PDL::ushort,PDL::long,PDL::longlong));

##-- see SF_FORMAT_PCM_* in sndfile.h
our %sfsubtype_to_ptype =
  (
   pcm_s8 => $s_nbits_to_ptype{8},		##-- signed 8 bit int
   pcm_u8 => $u_nbits_to_ptype{8},	        ##-- unsigned 8 bit int

   pcm_16  => $s_nbits_to_ptype{16},	##-- signed 16 bit int
   pcm_24 => $s_nbits_to_ptype{24},	##-- signed 24 bit int
   pcm_32 => $s_nbits_to_ptype{32},	##-- signed 32 bit int

   float => PDL::float,		##-- 32 bit floating point
   double => PDL::double,	##-- 64 bit floating point

   #ulaw => PDL::long,
   #alaw => PDL::long,
   #ima_adpcm => PDL::long,
   #ms_adpcm => PDL::long,
   #gsm610 => PDL::long,
   #vox_adpcm => PDL::long,
   #g721_32 => PDL::long,
   #g723_24 => PDL::long,
   #g723_40 => PDL::long,
   #dwvw_12 => PDL::long,
   #dwvw_16 => PDL::long,
   #dwvw_24 => PDL::long,
   #dwvw_n => PDL::long,

   dpcm_8 => $s_nbits_to_ptype{8},	##-- 8 bit differential PCM
   dpcm_16 => $s_nbits_to_ptype{16},	##-- 16 bit differential PCM

   DEFAULT => PDL::double,
  );

our %known_sf_subtypes =
  (map {$_=>undef}
   qw(pcm_s8 pcm_u8 pcm_16  pcm_24 pcm_32),
   qw(float double),
   qw(ulaw alaw ima_adpcm ms_adpcm gsm610 vox_adpcm),
   qw(g721_32 g723_24 g723_40),
   qw(dwvw_12 dwvw_16 dwvw_24 dwvw_n),
   qw(dpcm_8 dpcm_16),
  );

## $pdl_type = CLASS::_sfsubtype_to_ptype($sf_subtype)
sub _sfsubtype_to_ptype {
  my $sfsubtype = shift;
  my $ptype = $sfsubtype_to_ptype{$sfsubtype};
  return defined($ptype) ? $ptype : $sfsubtype_to_ptype{DEFAULT};
}

## $sf_subtype_or_undef = CLASS::_ptype_to_sfsubtype($pdl_type)
sub _ptype_to_sfsubtype {
  my $ptype = shift;
  $ptype = $ptype->type if (UNIVERSAL::can($ptype,'type'));
  my $nbits = 8*PDL::howbig($ptype);
  if ($ptype <= PDL::longlong) {
    ##-- integer types
    if ($nbits==8) { return 'pcm_u8'; }        ##-- unsigned 8 bit
    if ($nbits==16) { return 'pcm_16'; }       ##-- signed 16 bit
    if ($nbits==24) { return 'pcm_24'; }       ##-- signed 16 bit
    if ($nbits==32) { return 'pcm_32'; }       ##-- signed 32 bit
  }
  if ($nbits==32) { return 'float'; }          ##-- signed 32 bit floating point
  if ($nbits==64) { return 'double'; }         ##-- signed 64 bit floating point
  carp(__PACKAGE__."::_ptype_to_sfsubtype(): no default libsndfile subtype for PDL::Type '$ptype'");
  return undef;
}

## $sf_method_suffix = CLASS::_ptype_to_sfmethod($pdl_type)
sub _ptype_to_sfmethod {
  my $ptype = shift;
  $ptype = $ptype->type if (ref($ptype) && UNIVERSAL::can($ptype,'type'));
  #if ($ptype == PDL::byte) { return 'short'; } ##-- hack
  if ($ptype == PDL::short) { return 'short'; }
  if ($ptype == PDL::ushort) { return 'short'; }
  if ($ptype == PDL::long) { return 'int'; }
  #if ($ptype == PDL::ulong) { return 'int'; }
  #if ($ptype == PDL::longlong) { return 'int'; }
  if ($ptype == PDL::float) { return 'float'; }
  if ($ptype == PDL::double) { return 'double'; }
  return 'double'; ##-- default
}

##======================================================================
## Top-Level: I/O

##~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
## $pdl         = CLASS_OR_PDL->readsf($filename, %options); ##-- scalar context
## ($pdl,$info) = CLASS_OR_PDL->readsf($filename, %options); ##-- list context
##  + additional %options:
##     ptype  => $pdltype,
##     method => $readf_method_suffix,
##     channels, frames, ...
sub readsf {
  my ($that,$filename,%opts) = @_;

  ##-- open audio file
  my $sfio = PDL::IO::SndFile->new();
  $sfio->open("<",$filename,%opts)
    or confess(__PACKAGE__ . "::readsf(): open failed for file '$filename': $!");
  my $sf = $sfio->{sf};
  my $info = $sfio->info();
  %opts = (%$info, %opts);

  ##-- get pdl and its type
  my ($pdl);
  if (ref($that) && UNIVERSAL::isa($that,'PDL')) {
    $pdl = $that;
    $opts{ptype} = $pdl->type if (!defined($opts{ptype}));
  } elsif (ref($that)) {
    confess(__PACKAGE__ . "::readsf(): can't handle non-PDL reference as 1st arg");
  } else {
    $opts{ptype} = _sfsubtype_to_ptype($info->{subtype})
      if (!defined($opts{ptype})); ##-- guess PDL type from sf type
    $pdl = null->convert($opts{ptype});
  }

  ##-- reshape pdl if required
  if ($pdl->ndims != 2 || $pdl->dim(0) != $opts{channels} || $pdl->dim(1) != $opts{frames}) {
    $pdl = $pdl->reshape(@opts{qw(channels frames)});
  }

  ##-- ensure pdl is phyiscal
  $pdl = $pdl->make_physical;

  ##-- get sf read method suffix for pdl-type
  $opts{method} = _ptype_to_sfmethod($opts{ptype}) if (!$opts{method});
  my $readf_sub = $sf->can("readf_".$opts{method})
    or confess(__PACKAGE__."::readsf(): unknown SndFile readf_ suffix '$opts{method}'!");
  $readf_sub->($sf, ${$pdl->get_dataref}, $opts{frames});
  $pdl->upd_data();

  ##-- return
  return wantarray ? ($pdl,$info) : $pdl;
}

##~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
## $pdl         = $pdl->writesf($filename, %options); ##-- scalar context
## ($pdl,$info) = $pdl->writesf($filename, %options); ##-- list context
##  + additional %options:
##     ptype  => $pdl_type,                ##-- override $pdl->type
##     method => $writef_method_suffix,    ##-- override default for $pdl_type
##     samplerate => $sr,                  ##-- override package-global default
##     type => $sf_filetype,               ##-- override default file type (from filename or package-global)
##     subtype =>sf_subtype,               ##-- override default for $pdl_type
##     quiet => $bool,                     ##-- don't warn about implicit clumping
##     channels, frames, ...
sub writesf {
  my ($pdl,$filename,%opts) = @_;

  ##-- Sanity check(s)
  confess(__PACKAGE__ . "::writesf(): can't handle non-PDL passed as 1st argument")
    if (!ref($pdl) || !UNIVERSAL::isa($pdl,'PDL'));

  ##-- Options: Defaults: from %DEFAULTS
  $opts{samplerate} = $DEFAULTS{samplerate} if (!defined($opts{samplerate}));

  ##-- Options: Defaults: from filename or %DEFAULTS: type
  if (!defined($opts{type})) {
    if ($filename =~ /\.([^\.]+)$/) {
      $opts{type} = lc($1);
      delete($opts{type}) if (!exists($known_sf_file_types{$opts{type}}));
    }
    $opts{type} = $DEFAULTS{type} if (!$opts{type});
  }

  ##-- Options: Defaults: from PDL: channels, frames
  if ($pdl->ndims == 2) {
    $opts{channels} = $pdl->dim(0) if (!defined($opts{channels}));
    $opts{frames}   = $pdl->dim(1) if (!defined($opts{frames}));
  }
  elsif ($pdl->ndims < 2) {
    $opts{channels} = 1 if (!defined($opts{channels}));
    $opts{frames}   = $pdl->dim(0) if (!defined($opts{frames}));
  }
  else { ##-- ($pdl->ndims > 2)
    carp(__PACKAGE__."::writesf(): implicitly clumping to 2 dimensions") if (!$opts{quiet});
    $pdl = $pdl->clump(-2);
    $opts{channels} = $pdl->dim(0) if (!defined($opts{channels}));
    $opts{frames}   = $pdl->dim(1) if (!defined($opts{frames}));
  }

  ##-- Options: Defaults: from PDL: subtype & method
  $opts{ptype}   = $pdl->type if (!defined($opts{ptype}));
  $opts{subtype} = _ptype_to_sfsubtype($opts{ptype}) if (!$opts{subtype});
  $opts{method}  = _ptype_to_sfmethod($opts{ptype}) if (!$opts{method});

  ##-- Audio::SndFile object
  my $sfio = PDL::IO::SndFile->new();
  $sfio->open(">",$filename,%opts)
    or confess(__PACKAGE__ . "::writesf(): open failed for file '$filename'");
  my $sf = $sfio->{sf};

  ##-- Guts
  $pdl = $pdl->make_physical;
  my $writef_sub = $sf->can("writef_".$opts{method})
    or confess(__PACKAGE__."::writesf(): unknown SndFile writef_ suffix '$opts{method}'!");
  $writef_sub->($sf, ${$pdl->get_dataref});

  ##-- Return
  return wantarray ? ($pdl,$sfio->info) : $pdl;
}


##======================================================================
## Aliases
BEGIN {
  *PDL::readsf  = \&readsf;
  *PDL::writesf = \&writesf;
}

1; ##-- be happy

__END__

##========================================================================
## POD DOCUMENTATION, auto-generated by podextract.perl
=pod

=cut

##========================================================================
## NAME
=pod

=head1 NAME

PDL::IO::SndFile - PDL I/O routines for audio file formats supported by Audio::SndFile

=cut

##========================================================================
## SYNOPSIS
=pod

=head1 SYNOPSIS

 use PDL;
 use PDL::IO::SndFile;

 ($pdl,$info) = PDL->readsf($filename, %options);
 ($pdl,$info) = $pdl->writesf($filename, %options);

=cut

##========================================================================
## DESCRIPTION
=pod

=head1 DESCRIPTION

=cut

##----------------------------------------------------------------
## DESCRIPTION: PDL::IO::SndFile: Globals
=pod

=head2 Globals

=over 4

=item Variable: $VERSION

(undocumented)

=item Variable: %DEFAULTS

(undocumented)

=item Variable: %known_sf_file_types

(undocumented)

=item Variable: %known_sf_options

(undocumented)

=back

=cut

##----------------------------------------------------------------
## DESCRIPTION: PDL::IO::SndFile: Top-Level: I/O: Read
=pod

=head2 Top-Level: I/O

=over 4

=item readsf

 $pdl         = CLASS_OR_PDL->readsf($filename, %options);     ##-- scalar context;
 ($pdl,$info) = CLASS_OR_PDL-E<gt>readsf($filename, %options); ##-- list context

Read data from an audio file into a PDL.
Returned PDL will have audio channels along its 0th dimension and audio frames
along its 1st dimension.  If called in list context, the return value $info will
be a HASH-ref of audio file information as returned by Audio::SndFile.

=over 4

=item *

%options:

 ptype  => $pdltype,             ##-- output PDL type (default: guess from file)
 method => $readf_method_suffix, ##-- method suffix to use for reading raw data

... other options will be passed to Audio::SndFile::open().

=back

=item writesf

 $pdl         = $pdl->writesf($filename, %options); ##-- scalar context;
 ($pdl,$info) = $pdl->writesf($filename, %options); ##-- list context

Write PDL data to an audio file.
The calling $pdl should be max 2-dimensional,
having audio channels along its 0th dimension and audio frames
along its 1st dimension.  1d piddles are interpreted written as 1-channel (mono) audio files.
E<gt>2d piddles will be implicitly clump()ed down to 2 dimensions.
If called in list context, the return value $info will
be a HASH-ref of audio file information as returned by Audio::SndFile.

=over 4

=item *

known %options:

 ptype      => $pdl_type,                ##-- override $pdl->type
 method     => $writef_method_suffix,    ##-- override default for $pdl_type
 samplerate => $sr,                      ##-- override package-global default
 type       => $sf_filetyp               ##-- override default file type (from filename or package-global)
 subtype    => sf_subtype,               ##-- override default for $pdl_type
 quiet      => $bool,                    ##-- don't warn about implicit clumping

... other options will be passed to Audio::SndFile::open().

=back

=back

=cut

##----------------------------------------------------------------
## DESCRIPTION: PDL::IO::SndFile: Constructors etc.
=pod

=head2 Low-Level: Constructors etc.

=over 4

=item new

 $sfio = PDL::IO::SndFile->new(%args);


=over 4


=item *

%args:

 sf      => $sf,       ##-- underlying Audio::SndFile object
 options => \%options, ##-- options for Audio::SndFile::open()
 info    => \%info,    ##-- extended info for Audio::SndFile

=back

=item clear

 $sfio = $sfio->clear();


=over 4


=item *

clears cached data and options

=back

=back

=cut

##----------------------------------------------------------------
## DESCRIPTION: PDL::IO::SndFile: Methods: Open/close
=pod

=head2 Methods: Open/close

=over 4

=item open

 $sfio_or_undef = $sfio->open($mode,$file,%opts);
 $sfio_or_undef = $sfio->open("${mode}${file}",%opts)

=item close

 $sfio_or_undef = $sfio->close();

(undocumented)

=back

=cut

##----------------------------------------------------------------
## DESCRIPTION: PDL::IO::SndFile: Methods: Info
=pod

=head2 Methods: Info

=over 4

=item options

 \%options = $sfio->options();


=over 4


=item *

reads options from $sfio-E<gt>{sf}

=back

=item info

 \%info = $sfio->info();


=over 4


=item *

reads extended info from $sfio-E<gt>{sf}

=back

=back

=cut

##----------------------------------------------------------------
## DESCRIPTION: PDL::IO::SndFile: Utilities: dispatch to underlying Audio::SndFile object
=pod

=head2 Utilities: dispatch to underlying Audio::SndFile object

=over 4

=item Variable: %s_nbits_to_ptype

signed integer types: ($nbits=E<gt>$pdltype)

=item Variable: %u_nbits_to_ptype

unsigned integer types: ($nbits=E<gt>$pdltype)

=item Variable: %s_ptype_to_nbits

signed integer types: ($pdltype=E<gt>$nbits)

=item Variable: %u_ptype_to_nbits

unsigned integer types: ($pdltype=E<gt>$nbits)

=item Variable: %sfsubtype_to_ptype

(undocumented)

=item Variable: %known_sf_subtypes

(undocumented)

=item sfwrapper

 \&wrapped = CLASS::sfwrapper($func);

=item sfcall

 ? = $sfio->sfcall($func,@args);

=item _sfsubtype_to_ptype

 $pdl_type = CLASS::_sfsubtype_to_ptype($sf_subtype);

=item _ptype_to_sfsubtype

 $sf_subtype_or_undef = CLASS::_ptype_to_sfsubtype($pdl_type);

=item _ptype_to_sfmethod

 $sf_method_suffix = CLASS::_ptype_to_sfmethod($pdl_type);

=back

=cut

##========================================================================
## END POD DOCUMENTATION, auto-generated by podextract.perl
=pod

=head1 AUTHOR

Bryan Jurish E<lt>moocow@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009 by Bryan Jurish

This package is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.7 or,
at your option, any later version of Perl 5 you may have available.

=cut

